@@ -18,8 +18,8 @@
 #' @param nn_method Optional pre-calculated nearest neighbor data. There are
 #'   two supported formats. The first is a list consisting of two elements:
 #'   \itemize{
-#'     \item \code{"idx"}. A \code{n_vertices x n_neighbors} matrix where 
-#'     \code{n_vertices} is the number of observations in \code{X}. The contents 
+#'     \item \code{"idx"}. A \code{n_vertices x n_neighbors} matrix where
+#'     \code{n_vertices} is the number of observations in \code{X}. The contents
 #'     of the matrix should be the integer indexes of the data used to generate
 #'     the \code{model}, which are the \code{n_neighbors}-nearest neighbors of
 #'     the data to be transformed.
@@ -34,7 +34,7 @@
 #'   the \code{j}th observation in the original data used to generate the
 #'   \code{model} is a nearest neighbor of the \code{i}th observation in the new
 #'   data, with the distance given by the value of that element. In this format,
-#'   a different number of neighbors is allowed for each observation, i.e. 
+#'   a different number of neighbors is allowed for each observation, i.e.
 #'   each column can contain a different number of non-zero values.
 #'   Multiple nearest neighbor data (e.g. from two different pre-calculated
 #'   metrics) can be passed by passing a list containing the nearest neighbor
@@ -78,7 +78,7 @@
 #'     coordinates of the nearest neighbors from the original embedding in
 #'     \code{model}, where the weights used are the edge weights from the UMAP
 #'     smoothed knn distances. Equivalent to \code{init_weighted = TRUE}.
-#'     \item \code{"average"}. Use the mean average of the coordinates of 
+#'     \item \code{"average"}. Use the mean average of the coordinates of
 #'     the nearest neighbors from the original embedding in \code{model}.
 #'     Equivalent to \code{init_weighted = FALSE}.
 #'     \item A matrix of user-specified input coordinates, which must have
@@ -92,15 +92,15 @@
 #'   \code{learning_rate} and increase \code{n_epochs}, so whether this provides
 #'   a speed increase over the single-threaded optimization is likely to be
 #'   dataset and hardware-dependent. If \code{NULL}, the transform will use the
-#'   value provided in the \code{model}, if available. Default: \code{FALSE}. 
+#'   value provided in the \code{model}, if available. Default: \code{FALSE}.
 #' @param learning_rate Initial learning rate used in optimization of the
 #'   coordinates. This overrides the value associated with the \code{model}.
 #'   This should be left unspecified under most circumstances.
-#' @param opt_args A list of optimizer parameters, used when 
+#' @param opt_args A list of optimizer parameters, used when
 #'   \code{batch = TRUE}. The default optimization method used is Adam (Kingma
 #'   and Ba, 2014).
 #'   \itemize{
-#'     \item \code{method} The optimization method to use. Either \code{"adam"} 
+#'     \item \code{method} The optimization method to use. Either \code{"adam"}
 #'     or \code{"sgd"} (stochastic gradient descent). Default: \code{"adam"}.
 #'     \item \code{beta1} (Adam only). The weighting parameter for the
 #'     exponential moving average of the first moment estimator. Effectively the
@@ -119,28 +119,75 @@
 #'     step-size adaptivity and bring the behavior closer to stochastic gradient
 #'     descent with momentum. Typical values are between 1e-8 and 1e-3. Default:
 #'     \code{1e-7}.
-#'     \item \code{alpha} The initial learning rate. Default: the value of the 
+#'     \item \code{alpha} The initial learning rate. Default: the value of the
 #'     \code{learning_rate} parameter.
 #'   }
-#'   If \code{NULL}, the transform will use the value provided in the 
+#'   If \code{NULL}, the transform will use the value provided in the
 #'   \code{model}, if available.
 #' @param epoch_callback A function which will be invoked at the end of every
 #'   epoch. Its signature should be:
 #'   \code{(epoch, n_epochs, coords, fixed_coords)}, where:
 #'   \itemize{
-#'     \item \code{epoch} The current epoch number (between \code{1} and 
+#'     \item \code{epoch} The current epoch number (between \code{1} and
 #'     \code{n_epochs}).
 #'     \item \code{n_epochs} Number of epochs to use during the optimization of
 #'     the embedded coordinates.
 #'     \item \code{coords} The embedded coordinates as of the end of the current
 #'     epoch, as a matrix with dimensions (N, \code{n_components}).
 #'     \item \code{fixed_coords} The originally embedded coordinates from the
-#'     \code{model}. These are fixed and do not change. A matrix with dimensions 
+#'     \code{model}. These are fixed and do not change. A matrix with dimensions
 #'     (Nmodel, \code{n_components}) where \code{Nmodel} is the number of
 #'     observations in the original data.
 #'   }
+#' @param ret_extra A vector indicating what extra data to return. May contain
+#'   any combination of the following strings:
+#'   \itemize{
+#'     \item \code{"fgraph"} the high dimensional fuzzy graph (i.e. the fuzzy
+#'       simplicial set of the merged local views of the input data). The graph
+#'       is returned as a sparse matrix of class \link[Matrix]{dgCMatrix-class}
+#'       with dimensions \code{NX} x \code{Nmodel}, where \code{NX} is the number
+#'       of items in the data to transform in \code{X}, and \code{NModel} is
+#'       the number of items in the data used to build the UMAP \code{model}.
+#'       A non-zero entry (i, j) gives the membership strength of the edge
+#'       connecting the vertex representing the ith item in \code{X} to the
+#'       jth item in the data used to build the \code{model}. Note that the
+#'       graph is further sparsified by removing edges with sufficiently low
+#'       membership strength that they would not be sampled by the probabilistic
+#'       edge sampling employed for optimization and therefore the number of
+#'       non-zero elements in the matrix is dependent on \code{n_epochs}. If you
+#'       are only interested in the fuzzy input graph (e.g. for clustering),
+#'       setting \code{n_epochs = 0} will avoid any further sparsifying.
+#'   }
+#' @param seed Integer seed to use to initialize the random number generator
+#'   state. Combined with \code{n_sgd_threads = 1} or \code{batch = TRUE}, this
+#'   should give consistent output across multiple runs on a given installation.
+#'   Setting this value is equivalent to calling \code{\link[base]{set.seed}},
+#'   but it may be more convenient in some situations than having to call a
+#'   separate function. The default is to not set a seed, in which case this
+#'   function uses the behavior specified by the supplied \code{model}: If the
+#'   model specifies a seed, then the model seed will be used to seed then
+#'   random number generator, and results will still be consistent (if
+#'   \code{n_sgd_threads = 1}). If you want to force the seed to not be set,
+#'   even if it is set in \code{model}, set \code{seed = FALSE}.
 #' @return A matrix of coordinates for \code{X} transformed into the space
-#'   of the \code{model}.
+#'   of the \code{model}, or if \code{ret_extra} is specified, a list
+#'   containing:
+#'   \itemize{
+#'     \item \code{embedding} the matrix of optimized coordinates.
+#'     \item if \code{ret_extra} contains \code{"fgraph"}, an item of the same
+#'     name containing the high-dimensional fuzzy graph as a sparse matrix, of
+#'     type \link[Matrix]{dgCMatrix-class}.
+#'     \item if \code{ret_extra} contains \code{"sigma"}, returns a vector of
+#'     the smooth knn distance normalization terms for each observation as
+#'     \code{"sigma"} and a vector \code{"rho"} containing the largest
+#'     distance to the locally connected neighbors of each observation.
+#'     \item if \code{ret_extra} contains \code{"localr"}, an item of the same
+#'     name containing a vector of the estimated local radii, the sum of
+#'     \code{"sigma"} and \code{"rho"}.
+#'     \item if \code{ret_extra} contains \code{"nn"}, an item of the same name
+#'     containing the nearest neighbors of each item in \code{X} (with respect
+#'     to the items that created the \code{model}).
+#'   }
 #' @examples
 #'
 #' iris_train <- iris[1:100, ]
@@ -164,11 +211,19 @@ umap_transform <- function(X = NULL, model = NULL,
                            batch = NULL,
                            learning_rate = NULL,
                            opt_args = NULL,
-                           epoch_callback = NULL
-                           ) {
+                           epoch_callback = NULL,
+                           ret_extra = NULL,
+                           seed = NULL) {
   if (is.null(n_threads)) {
     n_threads <- default_num_threads()
   }
+  if (is.character(n_sgd_threads) && n_sgd_threads == "auto") {
+    n_sgd_threads <- n_threads
+  }
+  if (!is.numeric(n_sgd_threads)) {
+    stop("Unknown n_sgd_threads value: ", n_sgd_threads, " should be a positive
+         integer or 'auto'")
+  }
   if (is.null(nn_method)) {
     if (is.null(X)) {
       stop('argument "X" is missing, with no default')
@@ -177,41 +232,88 @@ umap_transform <- function(X = NULL, model = NULL,
       stop('argument "model" is missing, with no default')
     }
     if (!all_nn_indices_are_loaded(model)) {
-      stop("cannot use model: NN index is unloaded." ,
-           " Try reloading with `load_uwot`")
+      stop(
+        "cannot use model: NN index is unloaded.",
+        " Try reloading with `load_uwot`"
+      )
     }
-  }
-  else {
+  } else {
     if (!is.null(X)) {
       tsmessage('argument "nn_method" is provided, ignoring argument "X"')
       X <- NULL
     }
   }
-  
+
+  if (is.character(model$nn_method) &&
+      model$nn_method == "hnsw" && !is_installed("RcppHNSW")) {
+    stop(
+      "This model requires the RcppHNSW package to be installed."
+    )
+  }
+
+  if (is.character(model$nn_method) &&
+      model$nn_method == "nndescent" && !is_installed("rnndescent")) {
+    stop(
+      "This model requires the rnndescent package to be installed."
+    )
+  }
+
   if (is.null(n_epochs)) {
     n_epochs <- model$n_epochs
     if (is.null(n_epochs)) {
       if (ncol(graph) <= 10000) {
         n_epochs <- 100
-      }
-      else {
+      } else {
         n_epochs <- 30
       }
-    }
-    else {
+    } else {
       n_epochs <- max(2, round(n_epochs / 3))
     }
   }
-  
+
+  # Handle setting the random number seed internally:
+  # 1. If the user specifies seed = FALSE, definitely don't set the seed, even
+  # if the model has a seed.
+  # 2. If the user specifies seed = integer, then use that seed, even if the
+  # model has a seed.
+  # 3. If the user does not specify a seed, then use the model seed, if it
+  # exists. Otherwise don't set a seed. Also use this code path if the user
+  # sets seed = TRUE
+  if (is.logical(seed) && !seed) {
+    # do nothing
+  }
+  # handle the seed = TRUE case in this clause too
+  else if (is.logical(seed) || is.null(seed)) {
+    if (!is.null(model$seed)) {
+      tsmessage("Setting model random seed ", model$seed)
+      set.seed(model$seed)
+    }
+    # otherwise no model seed, so do nothing
+  } else {
+    tsmessage("Setting random seed ", seed)
+    set.seed(seed)
+  }
+
   if (is.null(search_k)) {
     search_k <- model$search_k
   }
-  
+
   nn_index <- model$nn_index
   n_neighbors <- model$n_neighbors
   local_connectivity <- model$local_connectivity
-  
+
   train_embedding <- model$embedding
+  if (!is.matrix(train_embedding)) {
+    # this should only happen if the user set
+    # `n_epochs = 0, init = NULL, ret_model = TRUE`
+    stop(
+      "Invalid embedding coordinates: should be a matrix, but got ",
+      paste0(class(train_embedding), collapse = " ")
+    )
+  }
+  if (any(is.na(train_embedding))) {
+    stop("Model embedding coordinates contains NA values")
+  }
   n_train_vertices <- nrow(train_embedding)
   ndim <- ncol(train_embedding)
   row.names(train_embedding) <- NULL
@@ -223,27 +325,25 @@ umap_transform <- function(X = NULL, model = NULL,
   metric <- model$metric
   nblocks <- length(metric)
   pca_models <- model$pca_models
-  
+
   if (method == "leopold") {
     dens_scale <- model$dens_scale
-    ai <- model$ai
+    aj <- model$ai
     rad_coeff <- model$rad_coeff
   }
-  
+
   if (is.null(batch)) {
     if (!is.null(model$batch)) {
       batch <- model$batch
-    }
-    else {
+    } else {
       batch <- FALSE
     }
   }
-  
+
   if (is.null(opt_args)) {
     if (!is.null(model$opt_args)) {
       opt_args <- model$opt_args
-    }
-    else {
+    } else {
       opt_args <- list()
     }
   }
@@ -253,11 +353,10 @@ umap_transform <- function(X = NULL, model = NULL,
   gamma <- model$gamma
   if (is.null(learning_rate)) {
     alpha <- model$alpha
-  }
-  else {
+  } else {
     alpha <- learning_rate
   }
-  if (! is.numeric(alpha) || length(alpha) > 1 || alpha < 0) {
+  if (!is.numeric(alpha) || length(alpha) > 1 || alpha < 0) {
     stop("learning rate should be a positive number, not ", alpha)
   }
   negative_sample_rate <- model$negative_sample_rate
@@ -269,12 +368,17 @@ umap_transform <- function(X = NULL, model = NULL,
     pcg_rand <- TRUE
   }
   num_precomputed_nns <- model$num_precomputed_nns
-  
+  binary_edge_weights <- model$binary_edge_weights
+  if (is.null(binary_edge_weights)) {
+    binary_edge_weights <- FALSE
+  }
+
   # the number of model vertices
   n_vertices <- NULL
   Xnames <- NULL
   if (!is.null(X)) {
-    if (!(methods::is(X, "data.frame") || methods::is(X, "matrix"))) {
+    if (!(methods::is(X, "data.frame") ||
+          methods::is(X, "matrix") || is_sparse_matrix(X))) {
       stop("Unknown input data format")
     }
     if (!is.null(norig_col) && ncol(X) != norig_col) {
@@ -296,31 +400,43 @@ umap_transform <- function(X = NULL, model = NULL,
     }
     checkna(X)
   } else if (nn_is_precomputed(nn_method)) {
+    # https://github.com/jlmelville/uwot/issues/97
+    # In the case where the training model didn't use pre-computed neighbors
+    # we treat it like it had one block
+    if (num_precomputed_nns == 0) {
+      num_precomputed_nns <- 1
+    }
     # store single nn graph as a one-item list
     if (num_precomputed_nns == 1 && nn_is_single(nn_method)) {
       nn_method <- list(nn_method)
     }
     if (length(nn_method) != num_precomputed_nns) {
-      stop("Expecting ", pluralize("pre-computed neighbor data block", num_precomputed_nns))
+      stop(
+        "Wrong # pre-computed neighbor data blocks, expected: ",
+        num_precomputed_nns, " but got: ", length(nn_method)
+      )
     }
     if (length(n_neighbors) != num_precomputed_nns) {
-      stop("Expecting ", 
-           pluralize("n_neighbor values (one per neighbor block)", num_precomputed_nns))
+      stop(
+        "Wrong # n_neighbor values (one per neighbor block), expected: ",
+        num_precomputed_nns, " but got: ", length(n_neighbors)
+      )
     }
     for (i in 1:num_precomputed_nns) {
       graph <- nn_method[[i]]
-      
+
       if (is.list(graph)) {
-        check_graph(graph, expected_rows = n_vertices, 
-                    expected_cols = n_neighbors[[i]], bipartite = TRUE)
+        check_graph(graph,
+          expected_rows = n_vertices,
+          expected_cols = n_neighbors[[i]], bipartite = TRUE
+        )
         if (is.null(n_vertices)) {
           n_vertices <- nrow(graph$idx)
         }
         if (is.null(Xnames)) {
           Xnames <- nn_graph_row_names(graph)
         }
-      }
-      else if (is_sparse_matrix(graph)) {
+      } else if (is_sparse_matrix(graph)) {
         # nn graph should have dims n_train_obs x n_test_obs
         graph <- Matrix::drop0(graph)
         if (is.null(n_vertices)) {
@@ -329,51 +445,50 @@ umap_transform <- function(X = NULL, model = NULL,
         if (is.null(Xnames)) {
           Xnames <- colnames(graph)
         }
-      }
-      else {
+      } else {
         stop("Error: unknown neighbor graph format")
       }
     }
     nblocks <- num_precomputed_nns
   }
-  
+
   if (!is.null(init)) {
     if (is.logical(init)) {
       init_weighted <- init
-    }
-    else if (is.character(init)) {
+    } else if (is.character(init)) {
       init <- tolower(init)
       if (init == "average") {
         init_weighted <- FALSE
-      }
-      else if (init == "weighted") {
+      } else if (init == "weighted") {
         init_weighted <- TRUE
-      }
-      else {
+      } else {
         stop("Unknown option for init: '", init, "'")
       }
-    }
-    else if (is.matrix(init)) {
+    } else if (is.matrix(init)) {
       indim <- dim(init)
       xdim <- c(n_vertices, ndim)
       if (!all(indim == xdim)) {
-        stop("Initial embedding matrix has wrong dimensions, expected (",
-             xdim[1], ", ", xdim[2], "), but was (",
-             indim[1], ", ", indim[2], ")")
+        stop(
+          "Initial embedding matrix has wrong dimensions, expected (",
+          xdim[1], ", ", xdim[2], "), but was (",
+          indim[1], ", ", indim[2], ")"
+        )
+      }
+      if (any(is.na(init))) {
+        stop("Initial embedding matrix coordinates contains NA values")
       }
       if (is.null(Xnames) && !is.null(row.names(init))) {
         Xnames <- row.names(init)
       }
       init_weighted <- NULL
-    }
-    else {
+    } else {
       stop("Invalid input format for 'init'")
     }
   }
   if (is.null(n_vertices)) {
     stop("Failed to read input correctly: invalid input format")
   }
-  
+
   if (verbose) {
     x_is_matrix <- methods::is(X, "matrix")
     tsmessage("Read ", n_vertices, " rows", appendLF = !x_is_matrix)
@@ -391,15 +506,22 @@ umap_transform <- function(X = NULL, model = NULL,
   graph <- NULL
   embedding <- NULL
   localr <- NULL
-  need_sigma <- method == "leopold" && nblocks == 1
+  sigma <- NULL
+  rho <- NULL
+  export_nns <- NULL
+  ret_nn <- FALSE
+  if ("nn" %in% ret_extra) {
+    ret_nn <- TRUE
+    export_nns <- list()
+  }
+  need_sigma <- (method == "leopold" && nblocks == 1) || "sigma" %in% ret_extra
   for (i in 1:nblocks) {
     tsmessage("Processing block ", i, " of ", nblocks)
     if (!is.null(X)) {
       if (nblocks == 1) {
         Xsub <- X
         ann <- nn_index
-      }
-      else {
+      } else {
         subset <- metric[[i]]
         if (is.list(subset)) {
           subset <- lsplit_unnamed(subset)$unnamed[[1]]
@@ -407,29 +529,86 @@ umap_transform <- function(X = NULL, model = NULL,
         Xsub <- X[, subset, drop = FALSE]
         ann <- nn_index[[i]]
       }
-      
+
       if (!is.null(pca_models) && !is.null(pca_models[[as.character(i)]])) {
         Xsub <- apply_pca(
           X = Xsub, pca_res = pca_models[[as.character(i)]],
           verbose = verbose
         )
       }
-      nn <- annoy_search(Xsub,
-                         k = n_neighbors, ann = ann, search_k = search_k,
-                         prep_data = TRUE,
-                         tmpdir = tmpdir,
-                         n_threads = n_threads, grain_size = grain_size,
-                         verbose = verbose
-      )
+      if (is.null(ann$type) || startsWith(ann$type, "annoy")) {
+        nn <- annoy_search(
+          Xsub,
+          k = n_neighbors,
+          ann = ann,
+          search_k = search_k,
+          prep_data = TRUE,
+          tmpdir = tmpdir,
+          n_threads = n_threads,
+          grain_size = grain_size,
+          verbose = verbose
+        )
+      }
+      else if (startsWith(ann$type, "hnsw")) {
+        if (is.list(model$nn_args)) {
+          nn_args <- model$nn_args
+          nn_args_names <- names(nn_args)
+
+          if ("ef" %in% nn_args_names) {
+            ef <- nn_args$ef
+          }
+          else {
+            ef <- 10
+          }
+        }
+
+        nn <- hnsw_search(
+          X,
+          k = n_neighbors,
+          ann = ann,
+          ef = ef,
+          n_threads = n_threads,
+          verbose = verbose
+        )
+
+        # We use the L2 HNSW index for Euclidean so we need to process the
+        # distances
+        if (names(model$metric)[[1]] == "euclidean") {
+          nn$dist <- sqrt(nn$dist)
+        }
+      }
+      else if (startsWith(ann$type, "nndescent")) {
+        nn <-
+          nndescent_search(
+            X,
+            k = n_neighbors,
+            ann = ann,
+            nn_args = model$nn_args,
+            n_threads = n_threads,
+            verbose = verbose
+          )
+      }
+      else {
+        stop("Unknown nn method: ", ann$type)
+      }
+      if (ret_nn) {
+        export_nns[[i]] <- nn
+        names(export_nns)[[i]] <- ann$metric
+      }
     } else if (is.list(nn_method)) {
       # otherwise we expect a list of NN graphs
       nn <- nn_method[[i]]
+      if (ret_nn) {
+        export_nns[[i]] <- nn
+        names(export_nns)[[i]] <- "precomputed"
+      }
+    } else {
+      stop(
+        "Can't transform new data if X is NULL ",
+        "and no sparse distance matrix available"
+      )
     }
-    else {
-      stop("Can't transform new data if X is NULL ", 
-           "and no sparse distance matrix available")
-    }
-    
+
     osparse <- NULL
     if (is_sparse_matrix(nn)) {
       nn <- Matrix::drop0(nn)
@@ -443,23 +622,23 @@ umap_transform <- function(X = NULL, model = NULL,
       }
       target <- log2(n_nbrs)
       skip_first <- TRUE
-    }
-    else {
+    } else {
       nnt <- nn_graph_t(nn)
       if (length(n_neighbors) == nblocks) {
         # if model came from multiple different external neighbor data
         n_nbrs <- n_neighbors[[i]]
-      }
-      else {
-        # multiple internal blocks 
+      } else {
+        # multiple internal blocks
         n_nbrs <- n_neighbors
       }
       if (is.na(n_nbrs) || n_nbrs != nrow(nnt$idx)) {
         # original neighbor data was sparse, but we are using dense knn format
         # or n_neighbors doesn't match
         n_nbrs <- nrow(nnt$idx)
-        tsmessage("Possible mismatch with original vs new neighbor data ", 
-                  "format, using ", n_nbrs, " nearest neighbors")
+        tsmessage(
+          "Possible mismatch with original vs new neighbor data ",
+          "format, using ", n_nbrs, " nearest neighbors"
+        )
       }
       target <- log2(n_nbrs)
       nn_ptr <- n_nbrs
@@ -482,17 +661,20 @@ umap_transform <- function(X = NULL, model = NULL,
     if (is.null(localr) && need_sigma) {
       # because of the adjusted local connectivity rho is too small compared
       # to that used to generate the "training" data but sigma is larger, so
-      # let's just stick with sigma + rho even though it tends to be an 
-      # underestimate 
+      # let's just stick with sigma + rho even though it tends to be an
+      # underestimate
+      sigma <- sknn_res$sigma
+      rho <- sknn_res$rho
       localr <- sknn_res$sigma + sknn_res$rho
     }
-    
+
     graph_blockv <- sknn_res$matrix
     if (is_sparse_matrix(nn)) {
-      graph_block <- Matrix::sparseMatrix(j = osparse$i, p = osparse$p, x = graph_blockv,
-                                dims = rev(osparse$dims), index1 = FALSE)
-    }
-    else {
+      graph_block <- Matrix::sparseMatrix(
+        j = osparse$i, p = osparse$p, x = graph_blockv,
+        dims = rev(osparse$dims), index1 = FALSE
+      )
+    } else {
       graph_block <- nn_to_sparse(nn_idxv, n_vertices, graph_blockv,
         self_nbr = FALSE,
         max_nbr_id = n_train_vertices,
@@ -514,17 +696,18 @@ umap_transform <- function(X = NULL, model = NULL,
         )
       if (is.null(embedding)) {
         embedding <- embedding_block
-      }
-      else {
+      } else {
         embedding <- embedding + embedding_block
       }
     }
-    
+
     if (is.null(graph)) {
       graph <- graph_block
-    }
-    else {
-      graph <- set_intersect(graph, graph_block, weight = 0.5, reset = FALSE)
+    } else {
+      graph <- set_intersect(graph, graph_block,
+        weight = 0.5,
+        reset_connectivity = FALSE
+      )
     }
   }
 
@@ -532,12 +715,21 @@ umap_transform <- function(X = NULL, model = NULL,
     if (nblocks > 1) {
       embedding <- embedding / nblocks
     }
-  }
-  else {
+  } else {
     tsmessage("Initializing from user-supplied matrix")
     embedding <- t(init)
   }
 
+  if (binary_edge_weights) {
+    tsmessage("Using binary edge weights")
+    graph@x <- rep(1, length(graph@x))
+  }
+
+  if (batch) {
+    # This is the same arrangement as Python UMAP
+    graph <- Matrix::t(graph)
+  }
+
   if (n_epochs > 0) {
     graph@x[graph@x < max(graph@x) / n_epochs] <- 0
     graph <- Matrix::drop0(graph)
@@ -548,33 +740,30 @@ umap_transform <- function(X = NULL, model = NULL,
     # i.e. the presence of (i->j) does NOT mean (j->i) is also present because
     # i and j now come from different data
     if (batch) {
-      # This is the same arrangement as Python UMAP
-      graph <- Matrix::t(graph)
-      # ordered indices of the new data nodes. Coordinates are updated 
+      # ordered indices of the new data nodes. Coordinates are updated
       # during optimization
       positive_head <- Matrix::which(graph != 0, arr.ind = TRUE)[, 2] - 1
       # unordered indices of the model nodes (some may not have any incoming
       # edges), these coordinates will NOT update during the optimization
       positive_tail <- graph@i
-    }
-    else {
-      # unordered indices of the new data nodes. Coordinates are updated 
+    } else {
+      # unordered indices of the new data nodes. Coordinates are updated
       # during optimization
       positive_head <- graph@i
       # ordered indices of the model nodes (some may not have any incoming edges)
       # these coordinates will NOT update during the optimization
       positive_tail <- Matrix::which(graph != 0, arr.ind = TRUE)[, 2] - 1
     }
-    
+
     n_head_vertices <- ncol(embedding)
     n_tail_vertices <- n_train_vertices
-    
+
     # if batch = TRUE points into the head (length == n_tail_vertices)
     # if batch = FALSE, points into the tail (length == n_head_vertices)
     positive_ptr <- graph@p
 
     epochs_per_sample <- make_epochs_per_sample(graph@x, n_epochs)
-    
+
     tsmessage(
       "Commencing optimization for ", n_epochs, " epochs, with ",
       length(positive_head), " positive edges",
@@ -584,23 +773,23 @@ umap_transform <- function(X = NULL, model = NULL,
     method <- tolower(method)
     if (method == "leopold") {
       # Use the linear model 2 log ai = -m log(localr) + c
-      aj <- exp(0.5 * ((-log(localr) * rad_coeff[2]) + rad_coeff[1]))
+      ai <- exp(0.5 * ((-log(localr) * rad_coeff[2]) + rad_coeff[1]))
       # Prevent too-small/large aj
-      min_aj <- min(sqrt(a * 10 ^ (-2 * dens_scale)), 0.1)
-      aj[aj < min_aj] <- min_aj
-      max_aj <- sqrt(a * 10 ^ (2 * dens_scale))
-      aj[aj > max_aj] <- max_aj
+      min_ai <- min(sqrt(a * 10^(-2 * dens_scale)), 0.1)
+      ai[ai < min_ai] <- min_ai
+      max_ai <- sqrt(a * 10^(2 * dens_scale))
+      ai[ai > max_ai] <- max_ai
       method <- "leopold2"
     }
 
-    method_args <- switch(method, 
+    method_args <- switch(method,
       umap = list(a = a, b = b, gamma = gamma, approx_pow = approx_pow),
       leopold2 = list(ai = ai, aj = aj, b = b, ndim = ndim),
       list()
     )
 
     full_opt_args <- get_opt_args(opt_args, alpha)
-    
+
     embedding <- optimize_layout_r(
       head_embedding = embedding,
       tail_embedding = train_embedding,
@@ -630,7 +819,28 @@ umap_transform <- function(X = NULL, model = NULL,
   if (!is.null(Xnames)) {
     row.names(embedding) <- Xnames
   }
-  embedding
+  if (length(ret_extra) > 0) {
+    res <- list(embedding = embedding)
+    for (name in ret_extra) {
+      if (name == "fgraph") {
+        res$fgraph <- graph
+      }
+      if (name == "sigma") {
+        res$sigma <- sigma
+        res$rho <- rho
+      }
+      if (name == "localr" && !is.null(localr)) {
+        res$localr <- localr
+      }
+      if (ret_nn && !is.null(export_nns)) {
+        res$nn <- export_nns
+      }
+    }
+  } else {
+    res <- embedding
+  }
+
+  res
 }
 
 init_new_embedding <-
@@ -677,8 +887,7 @@ init_transform <- function(train_embedding, nn_index, weights = NULL) {
       nbr_embedding <- train_embedding[nn_index[i, ], ]
       embedding[i, ] <- apply(nbr_embedding, 2, mean)
     }
-  }
-  else {
+  } else {
     for (i in 1:nr) {
       nbr_embedding <- train_embedding[nn_index[i, ], ]
       nbr_weights <- weights[nn_index[i, ], i]
@@ -699,18 +908,15 @@ apply_scaling <- function(X, scale_info, verbose = FALSE) {
     tsmessage("Applying training data range scaling")
     X <- X - scale_info[["scaled:range:min"]]
     X <- X / scale_info[["scaled:range:max"]]
-  }
-  else if (!is.null(scale_info[["scaled:maxabs"]])) {
+  } else if (!is.null(scale_info[["scaled:maxabs"]])) {
     tsmessage("Applying training data max-abs scaling")
     X <- scale(X, center = scale_info[["scaled:center"]], scale = FALSE)
     X <- X / scale_info[["scaled:maxabs"]]
-  }
-  else if (!is.null(scale_info[["scaled:colrange:min"]])) {
+  } else if (!is.null(scale_info[["scaled:colrange:min"]])) {
     tsmessage("Applying training data column range scaling")
     X <- sweep(X, 2, scale_info[["scaled:colrange:min"]])
     X <- sweep(X, 2, scale_info[["scaled:colrange:max"]], `/`)
-  }
-  else {
+  } else {
     tsmessage("Applying training data column filtering/scaling")
     X <- X[, scale_info[["scaled:nzvcols"]]]
     X <- scale(X,
@@ -729,20 +935,3 @@ apply_pca <- function(X, pca_res, verbose = FALSE) {
   }
   X %*% pca_res$rotation
 }
-
-all_nn_indices_are_loaded <- function(model) {
-  if (is.null(model$nn_index)) {
-    stop("Invalid model: has no 'nn_index'")
-  }
-  if (is.list(model$nn_index)) {
-    for (i in 1:length(model$nn_index)) {
-      if (model$nn_index$getNTrees() == 0) {
-        return(FALSE)
-      }
-    }
-  }
-  else if (model$nn_index$getNTrees() == 0) {
-    return(FALSE)
-  }
-  TRUE
-}
